#include "rtp_rx.h"
#include "g711.h"
#include "jbuf.h"
#include "ets_sys.h"
#include "osapi.h"
#include "user_interface.h"
#include <os_type.h>
#include <espconn.h>

//#define LOCAL_DEBUG
#ifdef LOCAL_DEBUG
#	define TRACE(args) (ets_uart_printf("RTP: "), ets_uart_printf args)
#else
#	define TRACE(args)
#endif
#	define MSG(args) (ets_uart_printf("RTP: "), printf args)

#define ntohs(x) ((((x) & 0xff) << 8) | (((x) & 0xff00) >> 8))
#define htonl(x) ( ((x)<<24 & 0xFF000000UL) | \
                   ((x)<< 8 & 0x00FF0000UL) | \
                   ((x)>> 8 & 0x0000FF00UL) | \
                   ((x)>>24 & 0x000000FFUL) )
#define ntohl(x) htonl(x)

enum {
	RTP_VERSION     =  2,  /**< Defines the RTP version we support */
	RTP_HEADER_SIZE = 12   /**< Number of bytes in RTP Header      */
};


/** Defines the RTP header */
struct rtp_header {
	uint8_t  ver;       /**< RTP version number     */
	bool     pad;       /**< Padding bit            */
	bool     ext;       /**< Extension bit          */
	uint8_t  cc;        /**< CSRC count             */
	bool     m;         /**< Marker bit             */
	uint8_t  pt;        /**< Payload type           */
	uint16_t seq;       /**< Sequence number        */
	uint32_t ts;        /**< Timestamp              */
	uint32_t ssrc;      /**< Synchronization source */
//	uint32_t csrc[16];  /**< Contributing sources   */
	struct {
		uint16_t type;  /**< Defined by profile     */
		uint16_t len;   /**< Number of 32-bit words */
	} x;
};

/**
    \return number of bytes occupied by header or value < 0 on error
*/
static int rtp_hdr_decode(struct rtp_header *hdr, const uint8_t *buf, int size)
{
	int err, i;
	size_t header_len;
	int left = size;

	if (left < RTP_HEADER_SIZE)
		return -1;

	hdr->ver  = (buf[0] >> 6) & 0x03;
	if (RTP_VERSION != hdr->ver)
		return -1;
	hdr->pad  = (buf[0] >> 5) & 0x01;
	hdr->ext  = (buf[0] >> 4) & 0x01;
	hdr->cc   = (buf[0] >> 0) & 0x0f;
	hdr->m    = (buf[1] >> 7) & 0x01;
	hdr->pt   = (buf[1] >> 0) & 0x7f;

	buf += 2;
	left -= 2;

	memcpy(&hdr->seq, buf, sizeof(hdr->seq));
	hdr->seq = ntohs(hdr->seq);
	buf += sizeof(hdr->seq);
	left -= sizeof(hdr->seq);
	memcpy(&hdr->ts, buf, sizeof(hdr->ts));
	hdr->ts = ntohl(hdr->ts);
	buf += sizeof(hdr->ts);
	left -= sizeof(hdr->ts);
	memcpy(&hdr->ssrc, buf, sizeof(hdr->ssrc));
	//ntohl
	buf += sizeof(hdr->ssrc);
	left -= sizeof(hdr->ssrc);

	header_len = hdr->cc*sizeof(uint32_t);
	if (left < header_len)
		return -2;

	for (i=0; i<hdr->cc; i++) {
		//hdr->csrc[i] = ntohl(mbuf_read_u32(mb));
		left -= sizeof(uint32_t);
		buf += sizeof(uint32_t);
	}

	if (hdr->ext) {
		if (left < 4)
			return -3;
		memcpy(&hdr->x.type, buf, sizeof(hdr->x.type));
		buf += sizeof(hdr->x.type);
		hdr->x.type = ntohs(hdr->x.type);
		memcpy(&hdr->x.len, buf, sizeof(hdr->x.len));
		buf += sizeof(hdr->x.len);
		hdr->x.len = ntohs(hdr->x.len);
		left -= 4;

		if (left < hdr->x.len*sizeof(uint32_t))
			return -4;

		buf += hdr->x.len*sizeof(uint32_t);
		left -= hdr->x.len*sizeof(uint32_t);
	}

	return size - left;
}


static void udpserver_recv(void *arg, char *pusrdata, unsigned short len) {
    /** \todo check source IP:port to block simultaneous transmissions */
//	unsigned long rcid = *(unsigned long*)pusrdata;
//	if( rcid == taskid ) got = 1;
//	struct espconn *pespconn = (struct espconn *)arg;
#if 0
    ets_uart_printf("RECV %p, %p, %dB\n", arg, pusrdata, (int)len);
    char buf[16];
    strncpy(buf, pusrdata, sizeof(buf));
    buf[sizeof(buf)-1] = '\0';
    ets_uart_printf("DATA: %s\n", buf);
#endif
    struct rtp_header hdr;
    int rc = rtp_hdr_decode(&hdr, pusrdata, len);
    if (len >= 0) {
        unsigned char *ptr;
        int i;
        TRACE(("RX pkt len %d, hdr size %d, pt %d, seq %d, ts %u\n", len, rc, hdr.pt, hdr.seq, hdr.ts));
        ptr = pusrdata + rc;
        if (hdr.pt == 8) {          // G.711a
            for (i=0; i<len-rc; i++) {
                int val = alaw2linear(*ptr);
                jbuf_put(val);
                ptr++;
            }
        } else if (hdr.pt == 0) {   // G.711u
            for (i=0; i<len-rc; i++) {
                int val = ulaw2linear(*ptr);
                jbuf_put(val);
                ptr++;
            }
        }
    } else {
        TRACE(("Error %d decoding, len %d\n", rc, len));
    }
    return;
}

static void ICACHE_FLASH_ATTR udp_init(void) {
    static struct espconn udpconn;
    static esp_udp udp;
    udpconn.type = ESPCONN_UDP;
    udpconn.state = ESPCONN_NONE;
    udpconn.proto.udp = &udp;
    udp.local_port = RTP_PORT;
    udpconn.reverse = NULL;
    espconn_regist_recvcb(&udpconn, udpserver_recv);
    int status = espconn_create(&udpconn);
    if (status) {
        ets_uart_printf("Failed to create UDP server, status = %d\n", status);
    } else {
        ets_uart_printf("UDP server created, listening on port %d\n", udp.local_port);
    }
}

void ICACHE_FLASH_ATTR rtp_rx_init(void) {
    udp_init();
}
